@@ -16,6 +16,7 @@ use std::error::Error;
 use std::fmt;
 use std::iter::FusedIterator;
 use std::str::FromStr;
+use thiserror::Error;
 
 #[cfg_attr(feature = "_serde", derive(Serialize, Deserialize))]
 #[derive(Clone, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
@@ -120,14 +121,14 @@ where
 declare_decimal_id! {
   pub struct DinoparcUserId(u32);
   pub type ParseError = DinoparcUserIdParseError;
-  const BOUNDS = 1..1_000_000_000;
+  const BOUNDS = 0..1_000_000_000;
   const SQL_NAME = "dinoparc_user_id";
 }
 
 declare_new_string! {
   pub struct DinoparcUsername(String);
   pub type ParseError = DinoparcUsernameParseError;
-  const PATTERN = r"^[0-9A-Za-z-]{1,14}$";
+  const PATTERN = r"^[0-9A-Za-z_-]{1,14}$";
   const SQL_NAME = "dinoparc_username";
 }
 
@@ -259,7 +260,7 @@ pub struct DinoparcUserIdRef {
 declare_decimal_id! {
   pub struct DinoparcDinozId(u32);
   pub type ParseError = DinoparcDinozIdParseError;
-  const BOUNDS = 1..1_000_000_000;
+  const BOUNDS = 0..1_000_000_000;
   const SQL_NAME = "dinoparc_dinoz_id";
 }
 
@@ -316,14 +317,33 @@ pub struct DinoparcDinozIdRef {
 #[cfg_attr(feature = "_serde", derive(Serialize, Deserialize))]
 #[cfg_attr(feature = "_serde", serde(tag = "type", rename = "DinoparcDinoz"))]
 #[derive(Clone, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
-pub struct ShortDinoparcDinoz {
+pub struct ShortDinoparcDinozWithLocation {
   pub server: DinoparcServer,
   pub id: DinoparcDinozId,
   pub name: DinoparcDinozName,
   pub location: DinoparcLocationId,
 }
 
-impl ShortDinoparcDinoz {
+impl ShortDinoparcDinozWithLocation {
+  pub const fn as_ref(&self) -> DinoparcDinozIdRef {
+    DinoparcDinozIdRef {
+      server: self.server,
+      id: self.id,
+    }
+  }
+}
+
+#[cfg_attr(feature = "_serde", derive(Serialize, Deserialize))]
+#[cfg_attr(feature = "_serde", serde(tag = "type", rename = "DinoparcDinoz"))]
+#[derive(Clone, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
+pub struct ShortDinoparcDinozWithLevel {
+  pub server: DinoparcServer,
+  pub id: DinoparcDinozId,
+  pub name: DinoparcDinozName,
+  pub level: u16,
+}
+
+impl ShortDinoparcDinozWithLevel {
   pub const fn as_ref(&self) -> DinoparcDinozIdRef {
     DinoparcDinozIdRef {
       server: self.server,
@@ -404,8 +424,8 @@ declare_new_enum!(
     Castivore,
     #[str("Gluon")]
     Gluon,
-    #[str("Gorriloz")]
-    Gorriloz,
+    #[str("Gorilloz")]
+    Gorilloz,
     #[str("Hippoclamp")]
     Hippoclamp,
     #[str("Kabuki")]
@@ -450,7 +470,7 @@ impl DinoparcDinozRace {
         '2' => Self::Castivore,
         '3' => Self::Sirain,
         '4' => Self::Winks,
-        '5' => Self::Gorriloz,
+        '5' => Self::Gorilloz,
         '6' => Self::Cargou,
         '7' => Self::Hippoclamp,
         '8' => Self::Rokky,
@@ -611,7 +631,7 @@ declare_new_enum!(
 pub struct DinoparcSessionUser<U = ShortDinoparcUser> {
   pub user: U,
   pub coins: u32,
-  pub dinoz: Vec<ShortDinoparcDinoz>,
+  pub dinoz: Vec<ShortDinoparcDinozWithLocation>,
 }
 
 #[cfg_attr(feature = "_serde", derive(Serialize, Deserialize))]
@@ -622,6 +642,16 @@ pub struct DinoparcInventoryResponse<U = ShortDinoparcUser> {
   pub inventory: HashMap<DinoparcItemId, u32>,
 }
 
+#[cfg_attr(feature = "_serde", derive(Serialize, Deserialize))]
+#[derive(Clone, Debug, PartialEq, Eq)]
+pub struct DinoparcExchangeWithResponse<U = ShortDinoparcUser> {
+  pub session_user: DinoparcSessionUser<U>,
+  pub own_bills: u32,
+  pub own_dinoz: Vec<ShortDinoparcDinozWithLevel>,
+  pub other_user: ShortDinoparcUser,
+  pub other_dinoz: Vec<ShortDinoparcDinozWithLevel>,
+}
+
 #[cfg_attr(feature = "_serde", derive(Serialize, Deserialize))]
 #[derive(Clone, Debug, PartialEq, Eq)]
 pub struct DinoparcCollectionResponse<U = ShortDinoparcUser> {
@@ -648,6 +678,11 @@ pub struct DinoparcDinozResponse<U = ShortDinoparcUser> {
 #[async_trait]
 #[auto_impl(&, Arc)]
 pub trait DinoparcClient: Send + Sync {
+  /// Returns the id of two distinct existing users on the provided Dinoparc server.
+  /// These users should be used when an `exchangeWith` query is required to fetch the
+  /// full Dinoz list but the specific target does not matter.
+  async fn get_preferred_exchange_with(&self, server: DinoparcServer) -> [DinoparcUserId; 2];
+
   async fn create_session(&self, options: &DinoparcCredentials) -> Result<DinoparcSession, EtwinError>;
 
   async fn test_session(
@@ -662,6 +697,12 @@ pub trait DinoparcClient: Send + Sync {
     id: DinoparcDinozId,
   ) -> Result<DinoparcDinozResponse, EtwinError>;
 
+  async fn get_exchange_with(
+    &self,
+    session: &DinoparcSession,
+    recipient: DinoparcUserId,
+  ) -> Result<DinoparcExchangeWithResponse, EtwinError>;
+
   async fn get_inventory(&self, session: &DinoparcSession) -> Result<DinoparcInventoryResponse, EtwinError>;
 
   async fn get_collection(&self, session: &DinoparcSession) -> Result<DinoparcCollectionResponse, EtwinError>;
@@ -678,11 +719,20 @@ pub trait DinoparcStore: Send + Sync {
 
   async fn touch_dinoz(&self, response: &DinoparcDinozResponse) -> Result<(), EtwinError>;
 
+  async fn touch_exchange_with(&self, response: &DinoparcExchangeWithResponse) -> Result<(), EtwinError>;
+
   async fn get_dinoz(&self, options: &GetDinoparcDinozOptions) -> Result<Option<ArchivedDinoparcDinoz>, EtwinError>;
 
   async fn get_user(&self, options: &GetDinoparcUserOptions) -> Result<Option<ArchivedDinoparcUser>, EtwinError>;
 }
 
+#[derive(Debug, Error)]
+pub enum GetExchangeWithError {
+  #[error("A user cannot exchange with itself")]
+  SelfExchange,
+}
+
+// TODO: Move to serde_tools
 #[cfg(feature = "_serde")]
 pub fn serialize_ordered_opt_temporal_map<K: Ord + Serialize, V: Serialize, S: Serializer>(
   value: &Option<LatestTemporal<HashMap<K, V>>>,
